/*   This is part of um-ViewOS
 *   The user-mode implementation of OSVIEW -- A Process with a View
 *
 *   viewfs.
 *   It is possible to remap files and directories
 *   
 *   Copyright 2005 Renzo Davoli University of Bologna - Italy
 *   
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *   $Id: viewfs.c,v 1.1.1.1 2005/05/19 15:36:32 garden Exp $
 *
 */   
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <utime.h>
#include <errno.h>
#include <sys/time.h>
#include <limits.h>
#include "module.h"
#include "libummod.h"

static char *homedir;
/* star stands for default situation */
/* per stands for default prersonality */
/* pom stands for plus or minus */
static char *starpath; 
static char *perpath;
static char *starpom;
static char *perpom;
static char stardir[2*PATH_MAX];
static char perdir[2*PATH_MAX];
static int critlen;

int read(), write(), close();

static int alwaysfalse()
{
	return 0;
}

static int viewfspath(char *path)
{
	if (*perdir == '/') {
		strcpy(path,perpath);
		*perpom='+';
		if (access(perdir,F_OK) == 0)
			return 1;
		*perpom='-';
		if (access(perdir,F_OK) == 0)
			return 1;
		if (errno == EACCES)
			return 1;
	}
	strcpy(path,starpath);
	*starpom='+';
	if (access(stardir,F_OK) == 0)
		return 1;
	*starpom='-';
	if (access(stardir,F_OK) == 0)
		return 1;
	if (errno == EACCES)
		return 1;
	return 0;
}

static char *unwrap(char *path)
{
	if (strncmp(path,stardir,critlen)==0)
		return NULL;
	if (*perdir == '/') {
		strcpy(path,perpath);
		*perpom='+';
		if (access(perdir,F_OK) == 0)
			return perdir;
		*perpom='-';
		if (access(perdir,F_OK) == 0)
			return NULL;
		if (errno == EACCES)
			return NULL;
	}
	strcpy(path,starpath);
	*starpom='+';
	if (access(stardir,F_OK) == 0)
		return stardir;
	*starpom='-';
	if (access(stardir,F_OK) == 0)
		return NULL;
	if (errno == EACCES)
		return NULL;
	return path;
}

static int viewfs_open(char *pathname, int flags, mode_t mode)
{
	return open(unwrap(pathname),flags,mode);
}

static int viewfs_stat(char *pathname, struct stat *buf)
{
	return stat(unwrap(pathname),buf);
}

static int viewfs_lstat(char *pathname, struct stat *buf)
{
	return lstat(unwrap(pathname),buf);
}

static int viewfs_stat64(char *pathname, struct stat64 *buf)
{
	return stat64(unwrap(pathname),buf);
}

static int viewfs_lstat64(char *pathname, struct stat64 *buf)
{
	return lstat64(unwrap(pathname),buf);
}

static int viewfs_readlink(char *path, char *buf, size_t bufsiz)
{
	return readlink(unwrap(path),buf,bufsiz);
}

static int viewfs_access(char *path, int mode)
{
	return access(unwrap(path),mode);
}

static int viewfs_mkdir(char *path, int mode)
{
	return mkdir(unwrap(path),mode);
}

static int viewfs_rmdir(char *path)
{
	return rmdir(unwrap(path));
}

static int viewfs_chmod(char *path, int mode)
{
	return chmod(unwrap(path),mode);
}

static int viewfs_chown(char *path, uid_t owner, gid_t group)
{
	return chown(unwrap(path),owner,group);
}

static int viewfs_lchown(char *path, uid_t owner, gid_t group)
{
	return lchown(unwrap(path),owner,group);
}

static int viewfs_unlink(char *path)
{
	return unlink(unwrap(path));
}

static int viewfs_link(char *oldpath, char *newpath)
{
	return link(unwrap(oldpath),unwrap(newpath));
}

static int viewfs_symlink(char *oldpath, char *newpath)
{
	return symlink(oldpath,unwrap(newpath));
}

static int viewfs_utime(char *filename, struct utimbuf *buf)
{
	return utime(unwrap(filename),buf);
}

static int viewfs_utimes(char *filename, struct timeval tv[2])
{
	return utimes(unwrap(filename),tv);
}

ssize_t viewfs_pread(int fd, void *buf, size_t count, long long offset)
{
	off_t off=offset;
	return pread(fd,buf,count,off);
}

ssize_t viewfs_pwrite(int fd, const void *buf, size_t count, long long offset)
{
	off_t off=offset;
	return pwrite(fd,buf,count,off);
}

static struct service s;

static void
__attribute__ ((constructor))
init (void)
{
	int len;
	printf("viewfs init\n");
	s.name="viewfs Virtual FS";
	s.code=0xf5;
	s.checkpath=viewfspath;
	s.checksocket=alwaysfalse;
	homedir=getenv("HOME");
	strcpy(stardir,homedir);
	strcat(stardir,"/.umview/*/x/");
	critlen=strlen(homedir)+9;
	len=strlen(stardir);
  starpom=stardir+len-2;  
  starpath=stardir+len;  
  *perdir=0;
	perpom=perdir+1;
	perpath=NULL;
	s.syscall=(intfun *)malloc(scmap_scmapsize * sizeof(intfun));
	s.socket=(intfun *)malloc(scmap_sockmapsize * sizeof(intfun));
	s.syscall[uscno(__NR_open)]=viewfs_open;
	s.syscall[uscno(__NR_creat)]=viewfs_open; /*creat must me mapped onto open*/
	s.syscall[uscno(__NR_read)]=read;
	s.syscall[uscno(__NR_write)]=write;
	s.syscall[uscno(__NR_readv)]=readv;
	s.syscall[uscno(__NR_writev)]=writev;
	s.syscall[uscno(__NR_close)]=close;
	s.syscall[uscno(__NR_stat)]=viewfs_stat;
	s.syscall[uscno(__NR_lstat)]=viewfs_lstat;
	s.syscall[uscno(__NR_fstat)]=fstat;
	s.syscall[uscno(__NR_stat64)]=viewfs_stat64;
	s.syscall[uscno(__NR_lstat64)]=viewfs_lstat64;
	s.syscall[uscno(__NR_fstat64)]=fstat64;
	s.syscall[uscno(__NR_readlink)]=viewfs_readlink;
	s.syscall[uscno(__NR_getdents)]=getdents;
	s.syscall[uscno(__NR_getdents64)]=getdents64;
	s.syscall[uscno(__NR_access)]=viewfs_access;
	s.syscall[uscno(__NR_fcntl)]=fcntl32;
	s.syscall[uscno(__NR_fcntl64)]=fcntl64;
	s.syscall[uscno(__NR__llseek)]=_llseek;
	s.syscall[uscno(__NR_lseek)]= (intfun) lseek;
	s.syscall[uscno(__NR_mkdir)]=viewfs_mkdir;
	s.syscall[uscno(__NR_rmdir)]=viewfs_rmdir;
	s.syscall[uscno(__NR_chown)]=viewfs_chown;
	s.syscall[uscno(__NR_lchown)]=viewfs_lchown;
	s.syscall[uscno(__NR_fchown)]=fchown;
	s.syscall[uscno(__NR_chmod)]=viewfs_chmod;
	s.syscall[uscno(__NR_fchmod)]=fchmod;
	s.syscall[uscno(__NR_unlink)]=viewfs_unlink;
	s.syscall[uscno(__NR_fsync)]=fsync;
	s.syscall[uscno(__NR_fdatasync)]=fdatasync;
	s.syscall[uscno(__NR__newselect)]=select;
	s.syscall[uscno(__NR_link)]=viewfs_link;
	s.syscall[uscno(__NR_symlink)]=viewfs_symlink;
	s.syscall[uscno(__NR_pread64)]=viewfs_pread;
	s.syscall[uscno(__NR_pwrite64)]=viewfs_pwrite;
	s.syscall[uscno(__NR_utime)]=viewfs_utime;
	s.syscall[uscno(__NR_utimes)]=viewfs_utimes;
	add_service(&s);
}

static void
__attribute__ ((destructor))
fini (void)
{
	free(s.syscall);
	free(s.socket);
	printf("viewfs fini\n");
}
